######################################################################
## INPUT                                                             #
## A polynomial f with power series coefficients and a non-negative  #
## integer r such that at least one coefficient of f is a unit.      #
## OUTPUT                                                            #
## Two polynomials p, alpha with power series coefficients such that #
## - the polynomial p is monic (p can be the constant one)           #
## - the polynomial alpha is invertible as a power series            #
## - we have f = p * alpha  mod  M^(r+1)                             #
######################################################################
TRDweierstrass_preparation := proc(upops::table, r::integer)
    local d, n, m, a, b, c, i;
    #################
    ## Checking in ##
    #################
    n := upops[DEG];
    a := upops[DATA];
    d := n+1;
    for i from 1 to (n+1) do
        if PS:-IsUnit(a[i]) then
            d := i-1;
            break;
        end if;
    end do;
    if (d = n+1) then error "Invalid input"; end if;
    if (d = 0) then return (TRDmake_upops_from_polynomial(1), upops); end if;
    ####################
    ## Initialization ##
    ####################
    b := Array([seq(PS:-Zero,i=0..d)]);
    b[d+1] := PS:-One;
    m := n - d;
    c := Array([seq(PS:-One,i=0..m)]);
    for i from 1 to (m+1) do
        c[i] := PS:-PolynomialToPowerSeries(PS:-HomegeneousPart(a[d+i],0));
    end do;
    #####################
    ## Iteration steps ##
    #####################
    for i from 1 to r do
        TRDweierstrass_iteration_step(a, b, c, i, n, d, m);
    end do;
    ##################
    ## Checking out ##
    ##################
    return (TRDmake_upops_from_array(b), TRDmake_upops_from_array(c));
end proc;

##################################################################
## INPUT                                                         #
## The arrays a,b,c are the dense representations of univariate  #
## polynomials f, p mod M^r, alpha mod M^r where                 #
## - M is the maximal ideal of the power series ring             #
## and such that                                                 #
## - the respective degrees of f, p, alpha are n,d,m             #
## - hence the arrays  a,b,c have n+1, d+1, m+1 elements         #
##   which are power series                                      #
## - we have f = p * alpha  mod  M^r                             #
## - the polynomial p is monic, hence b[d+1] is 1                #
## - the polynomial alpha is invertible as a power series,       #
##   hence c[1] (the constant term of alpha) is not zero         #
## OUTPUT                                                        #
## The arrays b,c have been updated such that all the above      #
## still hold; moreover we have f = p * alpha  mod  M^(r+1)      #
##################################################################

TRDweierstrass_iteration_step := proc(a::Array, b::Array, c::Array, r::integer, n::integer, d::integer, m::integer)
    local i, j, k, s;
    ################
    ## Updating b ##
    ################
    for i from 1 to d do
        s := a[i];
        for j from 2 to i do
           s := PS:-Subtract(s, PS:-Multiply(b[i-j+1], c[j]));
        end do;
        ## i=1 => s = a[1]
        ## i=2 => s = a[2] - b[1] * c[2]
        ## i=3 => s = a[3] - b[2] * c[2] - b[1] * c[3]
        b[i] := PS:-ExactQuotient(s,c[1]);
        b[i] := PS:-PolynomialToPowerSeries(PS:-PolynomialPart(b[i],r));
    end do;
    ################
    ## Updating c ##
    ################
    for i from 1 to (m+1) do
        s := a[d+i];
        for j from 1 to d do
            ## we must have (j-1) + (k-1) = d+i-1
            k := d+i+1 - j;
            if k <= m+1 then
               s := PS:-Subtract(s, PS:-Multiply(b[j], c[k]));
            end if;
        end do;
        ## i=1 => s = a[d+1] - b[1]*c[d+1] - ... - b[d]*c[2]
        ##  ...............................................................
        ## i=(m+1) => s = a[d+m+1] - b[1]*c[d+m+1] - ... - b[d]*c[m+2]
        c[i] := PS:-PolynomialToPowerSeries(PS:-PolynomialPart(s,r));
    end do;
    return (b, c);
end proc;

TRDsum := proc(f::table, g::table)
    local d_f, a_f, d_g, a_g, d_s, a_s, i;
    d_f := f[DEG];
    a_f := f[DATA];
    d_g := g[DEG];
    a_g := g[DATA];
    d_s := max(d_f,d_g);
    a_s := Array([seq(PS:-Zero,i=0..d_s)]);
    for i from 0 to d_s do
        if (i <= d_f) then
            a_s[i+1] := a_f[i+1];
        end if;
        if (i <= d_g) then
            a_s[i+1] := PS:-Add(a_s[i+1],a_g[i+1]);
        end if;
    end do;
    return TRDmake_upops_from_array(a_s);
end proc;

TRDsub := proc(f::table, g::table)
    local d_f, a_f, d_g, a_g, d_s, a_s, i;
    d_f := f[DEG];
    a_f := f[DATA];
    d_g := g[DEG];
    a_g := g[DATA];
    d_s := max(d_f,d_g);
    a_s := Array([seq(PS:-Zero,i=0..d_s)]);
    for i from 0 to d_s do
        if (i <= d_f) then
            a_s[i+1] := a_f[i+1];
        end if;
        if (i <= d_g) then
            a_s[i+1] := PS:-Subtract(a_s[i+1],a_g[i+1]);
        end if;
    end do;
    return TRDmake_upops_from_array(a_s);
end proc;

TRDproduct :=proc(f::table, g::table)
    local d_f, a_f, d_g, a_g, d_p, a_p, i, j;
    d_f := f[DEG];
    a_f := f[DATA];
    d_g := g[DEG];
    a_g := g[DATA];
    d_p := d_f + d_g;
    a_p := Array([seq(PS:-Zero,i=0..d_p)]);
    for i from 0 to d_f do
        for j from 0 to d_g do
            a_p[i+j+1] := PS:-Add(a_p[i+j+1], PS:-Multiply(a_f[i+1], a_g[j+1]));
        end do;
    end do;
    return TRDmake_upops_from_array(a_p);
end proc;

TRDpolynomial_part := proc(f::table, d::integer)
    local a_f, i, pol;
    a_f := f[DATA];
    pol := 0;
    for i from 1 to ArrayNumElems(a_f) do
        pol := pol + PS:-PolynomialPart(a_f[i],d) * Y^(i-1);
    end do;
    return (pol);
end proc;
